/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libs.h"
#include "libs-module.h"
#include "os_dep.h"
#include "error.h"
#include "events.h"
#include "threads.h"

#ifndef gpmi_no_static
#include "mods-static.h"
#endif

gpmi_module *gpmi_get_current_module()
{
	gpmi_private_globals_t *priv;

	priv = gpmi_threads_get_private();
	return priv->current_module;
}

void gpmi_set_current_module(gpmi_module *module)
{
	gpmi_private_globals_t *priv;

	priv = gpmi_threads_get_private();
	priv->current_module = module;
}


/**
 * Lowlevel load a gpmi modules.
 */
int gpmi_load_module(gpmi_module *module, const char *library_name)
{
	int state;
	char *tmp;

	/* Clear a few fields to avoid invalid frees when this call fails. */
	module->params      = NULL;
	module->path        = NULL;
	module->proc_UnInit = NULL;

	/* Attempt to load the library */
	gpmi_private.path_next = gpmi_private.path_override == NULL ? gpmi_shared.path_modules : gpmi_private.path_override;
	module->library        = gpmi_load_library(&gpmi_private.path_next, library_name, &state, 0);

	/* Check if failed */
	if (module->library == NULL) {
		if (state < 0)
			return state;
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_load_module, library_name);
		return gpmi_err_load_library;
	}

	gpmi_event_open(module);

	tmp              = strdup(module->library->name);
	module->mod_data = NULL;
	module->path     = strdup(dirname(tmp));
	free(tmp);

	/* Resolve module routines */
	module->proc_Init        = gpmi_dlsym_lowlevel(module->library, "gpmi_mod_Initialize");
	module->proc_UnInit      = gpmi_dlsym_lowlevel(module->library, "gpmi_mod_Uninitialize");
	module->proc_RunTick     = gpmi_dlsym_lowlevel(module->library, "gpmi_mod_RunTick");
	module->proc_RunData     = gpmi_dlsym_lowlevel(module->library, "gpmi_mod_RunData");
	module->proc_Debug       = gpmi_dlsym_lowlevel(module->library, "gpmi_mod_Debug");

	/* Check optional entry points */
	if (module->proc_Init == NULL)
		gpmi_err_stack_push(gpmi_err_entry_not_found, gpmi_err_src_load_module, "Initialize");

	if (module->proc_UnInit == NULL)
		gpmi_err_stack_push(gpmi_err_entry_not_found, gpmi_err_src_load_module, "UnInitialize");

	if (module->proc_Debug == NULL)
		gpmi_err_stack_push(gpmi_err_entry_not_found, gpmi_err_src_load_module, "Debug");

	if (module->proc_RunTick == NULL)
		gpmi_err_stack_push(gpmi_err_entry_not_found, gpmi_err_src_load_module, "RunTick");

	if (module->proc_RunData == NULL)
		gpmi_err_stack_push(gpmi_err_entry_not_found, gpmi_err_src_load_module, "RunData");

	return 0;
}

/**
 * Load and initialize a module.
 */
gpmi_module *gpmi_mod_load(const char *library_name, const char *params)
{
	gpmi_module *module;

	module = malloc(sizeof(gpmi_module));
	memset(module, 0, sizeof(gpmi_module));

#ifndef gpmi_no_static
	/* Check static modules */
	if (gpmi_load_static_module(module, library_name) == 0) {
		/* We have a static module by this name, so we are going to use it */
	} else
#endif
	{
		/* Not a static module, try to open the library */
		if (gpmi_load_module(module, library_name) != 0) {
			/* Couldn't load the lib -> no module, quit */
			free(module);
			return NULL;
		}
	}

	/* Clone the params if needed */
	if (params != NULL)
		module->params = strdup(params);

	/* Make sure we know that we are the current module */
	gpmi_set_current_module(module);

	/* Call the module's init routine if present */
	if (module->proc_Init != NULL) {
		if (module->proc_Init(module, params) != 0) {
			gpmi_mod_unload(module);
			return NULL;
		}
	}

	return module;
}

/**
 * Uninitialize (and unload) a module.
 */
void gpmi_mod_unload(gpmi_module *module)
{
	/* Make sure we know that we are the current module */
	gpmi_set_current_module(module);

	/* Call the UnInit of the module */
	if (module->proc_UnInit != NULL)
		module->proc_UnInit(module);

	/* Unload the underlaying lib if any (static modules don't have this) */
	if (module->library != NULL)
		gpmi_unload_library(module->library, NULL);

	/* Free params if we cloned it */
	if (module->params != NULL)
		free(module->params);

	if (module->path != NULL)
		free(module->path);

	/* Close the event-system for the module */
	gpmi_event_close(module);

	/* Free the module */
	free(module);
}
