/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file packages.h
 * \brief The include-file that is needed in combination with WrapGen output-files.
 *
 * WrapGen will automaticly add this file in its output-files.
 *
 * !!! Do not include it from gpmi.h - it's only for package developers !!!
 */

#ifndef GPMI_GPMI_PACKAGE_H
#define GPMI_GPMI_PACKAGE_H

#include <gpmi_config.h>
#include <gpmi/libs.h>
#include <string.h>
#ifndef gpmi_no_static
#include <gpmi/pkgs-static.h>
#endif

#if defined(WIN32) && !defined(__CYGWIN__)
	/* XXX: _strtoi64 declaration is not in MinGW headers */
	/* should be removed when win32api is fixed           */
	__int64 _strtoi64(const char *nptr, char **endptr, int base);
#	define strtoll _strtoi64
#endif

/**
 * \brief An strdup variant that reproduces NULL as ""
 */
#define strdup_safe(s) ((s != NULL) ? strdup(s) : strdup(""))

/**
 * \brief Register a package-command for the script-modules.
 */
void PkgRegisterCommand(void *handle, char *name, void *proc);

/**
 * \brief Macro to checks the number of arguments.
 */
#define PKG_Check_Argc(paramnum)	\
	if ( (argc-1) != paramnum ) {		\
		gpmi_error(0, "[Pkg] Error, wrong number of arguments at line %s:%d in %s: got %d expected %d\n", __FILE__, __LINE__, __FUNCTION__, argc-1, paramnum); \
		return NULL;									\
	}

/**
 * \brief Returns different kind of arguments as a dynamic string.
 */
#define PKG_Result_String(str) Result = str;

/**
 * \brief Return keyword as string.
 */
#define PKG_Result_Keyword(array, res)									\
	Result = (char *)gpmi_keyword_find_rev(array, (res));	\
	return strdup_safe(Result);

/**
 * \brief Return multiple-keywords as string.
 */
#define PKG_Result_Keyword_Multiple(array, res)								\
	Result = (char *)gpmi_keyword_find_rev_multi(array, (res)); \
	if (Result == NULL) 																				\
		Result = strdup("");																			\
	return Result

/**
 * \brief Return integer as string.
 */
#define PKG_Result_Int(res) 		Result = malloc(32); 													\
																sprintf(Result, "%d", (int)(res));

/**
 * \brief Return unsigned integer as string.
 */
#define PKG_Result_Uint(res)		Result = malloc(32); 													\
																sprintf(Result, "%u", (unsigned int)(res));

/**
 * \brief Return double as string.
 */
#define PKG_Result_Double(res)  Result = malloc(64); 													\
																sprintf(Result, "%g", (double)(res));

/**
 * \brief Call function and return NULL.
 */
/* 'res;' is very much needed, DO NOT REMOVE! */
#define PKG_Result_Void(res)		res; Result = NULL;

/**
 * \brief Return pointer as string.
 */
#define PKG_Result_Ptr(res) 		Result = malloc(32); 													\
																sprintf(Result, PRINTF_PTR, STD_PRINTFP_CAST(void *)(res));


/**
 * \brief Macros to convert parameters with error handling for the function wrappers.
 *  They will return on error so don't have unfreed memory before the call
 */
#define PKG_KeywordErrorAndReturn(result, param_no) \
	gpmi_error(0, "[PKG] Error, invalid single keyword argument %d at function %s: %s\n", param_no, __FUNCTION__, argv[param_no]); \
	return NULL;

#define PKG_ConvertInteger(result, param_no) \
	{ \
		char *PKG_ConvertInteger_tmp; \
		result = strtol(argv[param_no], &PKG_ConvertInteger_tmp, 0); \
		if ( *PKG_ConvertInteger_tmp != 0 ) { \
			gpmi_error(0, "[PKG] Error, invalid integer argument %d at function %s: %s\n", param_no, __FUNCTION__, argv[param_no]); \
			return NULL; \
		} \
	}

#define PKG_ConvertString(result, param_no) \
	result = argv[param_no];


#define PKG_ConvertPtr(result, param_no) \
	{ \
		char *PKG_ConvertInteger_tmp; \
		result = (void *)(size_t)strtoll(argv[param_no], &PKG_ConvertInteger_tmp, 16); \
		if ( *PKG_ConvertInteger_tmp != 0 ) { \
			gpmi_error(0, "[PKG] Error, invalid integer argument %d at function %s: %s\n", param_no, __FUNCTION__, argv[param_no]); \
			return NULL; \
		} \
	}

#define PKG_ConvertDouble(result, param_no) \
	{ \
		char *PKG_ConvertDouble_tmp; \
		result = (double)strtod(argv[param_no], &PKG_ConvertDouble_tmp); \
		if ( *PKG_ConvertDouble_tmp != 0 ) { \
			gpmi_error(0, "[PKG] Error, invalid double argument %d at function %s: %s\n", param_no, __FUNCTION__, argv[param_no]); \
			return NULL; \
		} \
	}


/* The following part is for listgen ------------------------------------ */

/* Defining a gpmi event in a package .h file.
		Syntax is:
			gpmi_define_event(name)(param1, param2, param3);

		gpmi_define_event creates the "int name" definition then calls
		gpmi_define_eatup that only eats up the param list in the second ().
*/

#ifndef NO_VARARG_MACRO
	/* Vararg macro version: eat up (ignore) parameters */
#	define gpmi_define_eatup(x...)
#	define gpmi_define_event(name)		int name gpmi_define_eatup

#	define FAKETYPE(x...)
#else
	/* No vararg macro version: define unique dummy functions so we can use the parameters */
#	define gpmi_define_event(name)		int name ; void name ## _novarargmacro_event	

#	define FAKETYPE void name ## _novarargmacro_faketype
#	define gpmi_fake_keyword void name ## _novarargmacro_fake_keyword
#endif

/* These prefixes are only for listgen parser, eat them up in C */
#define multiple
#define dynamic
#define direct
#define nowrap

#endif /* GPMI_GPMI_PACKAGE_H */
