/*
    libgpmi - General Package/Module Interface - timer packge
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * Timers for modules: the caller may set up timer lists. When a timer triggers,
 * the callback routine of the list is called with a timer-specific data.
 *
 * Features:
 *  - Timers can be dynamicallly added and deleted.
 *  - Timers have a TTL field that describes how many times they may run
 *     (after that, they are deleted automaticly)
 *  - A timer with Frequency=0 runs a cycle - this may speed up slow
 *     script interpreters
 *
 * Tips:
 *  - Data should be unique for each timer if you plan to delete timers
 *  - if TTL is 0, the timer runs forever
 *  - run a cycle: TTL=<nmber of runs>, Frequency=0
 *
 * Usage:
 *  Build a TimerList (with timer_init), and add timers in it
 *  (timer_add) call timer_run every tick, and that is all.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <gpmi.h>
#include "timer.h"

/* Helper functions and macros */

/* Update NextFrame if the given timer would trigger sooner than the original value */
#define TimerUpdateNext(TimerList, timer)		if ( TimerList->next_frame > timer->next_frame || TimerList->next_frame == 0 ) \
																								TimerList->next_frame = timer->next_frame;


/**
 * Call the proc to inform the timer should be destroyed
 */
static void timer_destroy(timer_list *TimerList, timer *Timer)
{
	/* Call the proc to clean up the timer's data */
	TimerList->CallBackProc(TIMER_UNINIT, Timer->data, Timer);
}

/**
 * Call the proc to inform it is time to tick
 */
static int timer_call(timer_list *TimerList, timer *Timer, int frame_counter)
{
	/* Call the proc to clean up the timer's data */
	TimerList->CallBackProc(TIMER_TICK, Timer->data, Timer);

	/* Decrease TTL, if TTL presents */
	if ( Timer->TTL > 0) {
		Timer->TTL--;
		if ( Timer->TTL == 0 )
			return 1;
	}

	/* Set up for the next triggering	*/
	Timer->next_frame = frame_counter + Timer->frequency;

	TimerUpdateNext(TimerList, Timer);

	return 0;
}

/* Interface functions */

/**
 * Initialize a TimerList
 *
 * @return Returns the TimerList
 */
timer_list *timer_init(timer_callback *CallBackProc)
{
	timer_list *TimerList;

	/* Allocate memory for the structure, make sure we have room */
	TimerList = malloc(sizeof(timer_list));
	assert( TimerList != NULL );

	/* Fill in initial data */
	TimerList->CallBackProc = CallBackProc;
	TimerList->next_frame   = 0;
	TimerList->FirstTimer   = NULL;

	/* Return the strucutre */
	return TimerList;
}

/**
 * Uninitialize a TimerList
 */
void timer_uninit(timer_list *TimerList)
{
	timer *current, *next;

	/* Free all the timers, if any */
	if ( TimerList->FirstTimer != NULL )  {
		/* Walk trough the list of timers and destroy each */
		for (current = TimerList->FirstTimer; current != NULL; current = next) {
			next = current->next;
			timer_destroy(TimerList, current);
			free(current);
		}
	}

	free(TimerList);
}

/**
 * Add a Timer to a TimerList
 */
timer *timer_add(timer_list *TimerList, int frame_counter, int TTL, int frequency, void *data)
{
	timer *tmr, *tmp;
	int n;

	/* Allocate and create the timer */
	tmr = malloc(sizeof(timer));
	tmr->frequency  = frequency;
	tmr->next_frame = frame_counter + frequency;
	tmr->TTL  = TTL;
	tmr->data = data;

	/* Call the proc to set up */
	TimerList->CallBackProc(TIMER_INIT, data, tmr);

	if ( frequency == 0 ) {
		/* Cycles - run the timer TTL times immediatly */
		for (n = TTL; n > 0; n--) {
			TimerList->CallBackProc(TIMER_TICK, data, tmr);
			/* Decrease the TTL */
			tmr->TTL--;
		}

		/* Clean up */
		timer_destroy(TimerList, tmr);
		free(tmr);
		tmr = NULL;
	}
	else {
		/* For real timers - link it in the timer list */

		/* We have to append here, insert won't work when one creates a timer
			 from a timer proc that is getting out of TTL.
			 Find the last timer on the list  */
		tmr->next = NULL;
		tmp = TimerList->FirstTimer;

		/* Is this our first timer? */
		if ( tmp == NULL ) {
			TimerList->FirstTimer = tmr;
		} else {
			/* Search end of the list */
			while ( tmp->next != NULL )
				tmp = tmp->next;

			/* append */
			tmp->next = tmr;
		}

		TimerUpdateNext(TimerList, tmr);
	}

	return tmr;
}

/**
 * Test all Timers from a TimerList to see if they need to be executed
 */
void timer_run(timer_list *TimerList, int frame_counter)
{
	timer *timer, *last_timer, *next_timer;

	/* Optimization: don't do anything if we don't need to */
	if ( TimerList->next_frame > frame_counter && TimerList->next_frame != 0 )
		return;

	/* So we have at least one timer that has to be run now.
		 Walk trough the linked list */
	last_timer = NULL;
	for (timer = TimerList->FirstTimer; timer != NULL; timer = next_timer) {
		next_timer = timer->next;

		if ( timer->next_frame <= frame_counter ) {
			/* Run this timer */
			if ( timer_call(TimerList, timer, frame_counter) ) {
				/* The timer is over, have to be killed: deleting from a linked list */
				if ( last_timer == NULL )
					TimerList->FirstTimer = timer->next;
				else
					last_timer->next = timer->next;

				/* Destroying the timer */
				timer_destroy(TimerList, timer);
				free(timer);
			} else {
				last_timer = timer;
			}
		}
		else {
			/* No run, still can be the next soonest-triggering timer so we update */
			TimerUpdateNext(TimerList, timer);
			last_timer = timer;
		}
	}

	/* No timers are left? */
	if ( TimerList->next_frame <= frame_counter )
		TimerList->next_frame = 0;
}


/**
 * Del a Timer from a TimerList
 */
int timer_del(timer_list *TimerList, timer *ptr)
{
	timer *t, *last;

	for (last = NULL, t = TimerList->FirstTimer; t != NULL; last = t, t = t->next) {
		/* Find the requested timer */
		if (t == ptr) {
			/* First on the list? */
			if (last != NULL) {
				/* Not the first */
				last->next = t->next;
			}
			else {
				/* First */
				TimerList->FirstTimer = t->next;
			}

			/* Free timer data */
			timer_destroy(TimerList, t);
			free(t);
			return 0;
		}
	}

	return -1;
}
